# Copyright (C) 2006-2020 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import time
import os
import subprocess
import logging

def run_cmd(args, success_return_code=(0,), logargs=None, stdin=None, run_background=False):
    logargs = logargs or args
    stdoutdata = ''
    if not run_background:
        proc = subprocess.Popen(args, stdin=stdin,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, 
                                close_fds=True)
        stdoutdata, stderrdata = proc.communicate()
        if proc.returncode not in success_return_code:
            raise Exception('Command: %s failed (%s): stderr: %s stdout: %s'
                            % (logargs, proc.returncode, stderrdata, stdoutdata))
    else:
        subprocess.Popen(args, stdin=stdin,
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return stdoutdata


def readfile(filepath):
    with open(filepath) as f:
        return f.read()


def get_all_vms():
    """Get all VMs through xm."""
    vms = []
    lines = run_cmd(['xm', 'list']).splitlines()[1:]
    for line in lines:
        vm_name = line.split()[0].strip()
        if vm_name != 'Domain-0':
            vms.append(vm_name)
    return vms


def get_all_vm_vif_macs():
    """Return a vif-mac table of all VMs in a list."""
    vms = get_all_vms()
    vif_mac_table = {}
    for vm in vms:
        lines = run_cmd(['xm', 'network-list', vm]).splitlines()[1:]
        for line in lines:
            network_info = line.split()
            node_name = 'backend' + network_info[8].strip().split('backend')[1]
            mac = network_info[2].strip()
            vif_mac_table[node_name] = mac
    return vif_mac_table


def get_bond_active_slave(bond):
    """Get an active-backup bond's current active slave."""
    return readfile('/sys/class/net/%s/bonding/active_slave' % bond).strip()


def get_active_backup_bonds():
    """Get all bonds that are in active-backup mode."""
    bonds = readfile('/sys/class/net/bonding_masters').strip().split()
    active_backup_bonds = []
    for bond in bonds:
        active_slave = get_bond_active_slave(bond)
        if bond != 'bond0' and active_slave != '' and \
        readfile('/sys/class/net/%s/bonding/mode' % bond).strip().split()[1] == '1':
            active_backup_bonds.append(bond)
    return active_backup_bonds


def get_bond_bridges_and_vlans(bond):
    """Get all bridges and vlans that are created on a bond."""
    upper_dirs = run_cmd(['find', '/sys/class/net/%s/' % bond, '-name', 'upper_*']).splitlines()
    bridges = []
    vlans = []
    for upper_dir in upper_dirs:
        upper_dir = upper_dir.strip()
        device_name = upper_dir.split('upper_', 1)[1]
        if os.path.isdir(upper_dir + '/bridge'):
            bridges.append(device_name)
        else:
            vlans.append(device_name)
    return bridges, vlans


def get_bridge_macs(br, bond, vm_vif_mac_table):
    """Look up all MAC addresss attaching to a bridge using the existing vif-mac table."""
    macs = []
    vif_dirs = run_cmd(['find','/sys/class/net/%s/' % br, '-name', 'lower_*']).splitlines()
    for vif_dir in vif_dirs:
        vif_dir = vif_dir.strip()
        vif_name = vif_dir.split('lower_', 1)[1]
        if vif_name != bond:
            if os.path.exists('%s/device/nodename' % vif_dir):
                nodename = readfile('%s/device/nodename' % vif_dir).strip()
                if nodename in vm_vif_mac_table:
                    mac = vm_vif_mac_table[nodename]
                    macs.append(mac)
    return macs


def send_frame_with_mac(mac, interface, count=3):
    """Use the arping tool to send out frames with fake MAC address through specific interface."""
    run_cmd(['/opt/ovs-extra/arping', '-B0', '-ps', mac, '-i', interface, '-c', str(count)], 
            success_return_code=(0,1), run_background=True)


def bond_slave_port_update(frame_count):
    """
       Store all bonds and their active slaves in a table and monitor every active slave change.
       If any bond active slave changes, first find all the bridges attaching to it and its vlans.
       And then find all the vif MAC addresses that are attaching to those bridges.
       Finally, update the network switch by sending frames with these source MAC addresses
       and update the bond slave table for the next time detection.
    """
    global bond_slave_table

    bonds = get_active_backup_bonds()

    for bond in bonds:
        active_slave = get_bond_active_slave(bond)

        if bond not in bond_slave_table:
            bond_slave_table[bond] = active_slave

        elif active_slave != bond_slave_table[bond]:
            
            logger.info("bond '%s' active slave changed from '%s' to '%s'" 
                        % (bond, bond_slave_table[bond], active_slave)) 
            brs, vlans = get_bond_bridges_and_vlans(bond)
            logger.info("bridges attached to bond '%s' are: %s" % (bond, brs))
            logger.info("vlans attached to bond '%s' are: %s" % (bond, vlans))
            vm_vif_mac_table = get_all_vm_vif_macs()

            for br in brs:
                macs = get_bridge_macs(br, bond, vm_vif_mac_table)
                logger.info("mac addresses attached to bridge '%s' are: %s" % (br, macs))
    
                for mac in macs:
                    send_frame_with_mac(mac, br, frame_count)

            for vlan in vlans:
                vlan_brs = get_bond_bridges_and_vlans(vlan)[0]
                for vlan_br in vlan_brs:
                    macs = get_bridge_macs(vlan_br, vlan, vm_vif_mac_table)
                    logger.info("mac addresses attached to vlan bridge '%s' are: %s" % (vlan_br, macs))

                    for mac in macs:
                        send_frame_with_mac(mac, vlan_br, frame_count)

            bond_slave_table[bond] = active_slave

service = 'ovs-arping'
bond_slave_table = {}
frame_count = 3		# 3 frames are sent for a single affacted mac address by default

logger = logging.getLogger('%s' % service)
logger.setLevel(logging.DEBUG)
handler = logging.FileHandler('/var/log/%s.log' % service)
handler.setLevel(logging.DEBUG)
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s', '%Y-%m-%d %H:%M:%S'))
logger.addHandler(handler)

if __name__ == '__main__':

    try:
        logger.info("%s service started.\n" % service)
        while True:
            bond_slave_port_update(frame_count)
            # every frame takes 1 second to be sent so 'frame_count' seconds are required for one MAC
            # the update frames for all MAC addresses are processed in parallel
            # so adding 1 second for all the rest of the program to complete before another round check
            time.sleep(frame_count + 1)
    except Exception as e:
        logger.error("ERROR: %s, please check the issue and restart %s service\n" % (str(e), service))
